@@ -1,27 +1,27 @@
-## Copyright (C) 2014-2015 Eduardo Ramos Fernández <eduradical951@gmail.com>
-## Copyright (C) 2013-2015 Kai T. Ohlhus <k.ohlhus@gmail.com>
+## Copyright (C) 2014-2018 Eduardo Ramos Fernández <eduradical951@gmail.com>
+## Copyright (C) 2013-2018 Kai T. Ohlhus <k.ohlhus@gmail.com>
 ##
 ## This file is part of Octave.
 ##
-## Octave is free software; you can redistribute it and/or modify it
+## Octave is free software: you can redistribute it and/or modify it
 ## under the terms of the GNU General Public License as published by
-## the Free Software Foundation; either version 3 of the License, or (at
-## your option) any later version.
+## the Free Software Foundation, either version 3 of the License, or
+## (at your option) any later version.
 ##
 ## Octave is distributed in the hope that it will be useful, but
 ## WITHOUT ANY WARRANTY; without even the implied warranty of
-## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-## General Public License for more details.
+## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+## GNU General Public License for more details.
 ##
 ## You should have received a copy of the GNU General Public License
 ## along with Octave; see the file COPYING.  If not, see
-## <http://www.gnu.org/licenses/>.
+## <https://www.gnu.org/licenses/>.
 
 ## -*- texinfo -*-
-## @deftypefn  {Function File} {} ilu (@var{A})
-## @deftypefnx {Function File} {} ilu (@var{A}, @var{opts})
-## @deftypefnx {Function File} {[@var{L}, @var{U}] =} ilu (@dots{})
-## @deftypefnx {Function File} {[@var{L}, @var{U}, @var{P}] =} ilu (@dots{})
+## @deftypefn  {} {} ilu (@var{A})
+## @deftypefnx {} {} ilu (@var{A}, @var{opts})
+## @deftypefnx {} {[@var{L}, @var{U}] =} ilu (@dots{})
+## @deftypefnx {} {[@var{L}, @var{U}, @var{P}] =} ilu (@dots{})
 ##
 ## Compute the incomplete LU factorization of the sparse square matrix @var{A}.
 ##
@@ -42,7 +42,7 @@
 ## Type of factorization.
 ##
 ## @table @asis
-## @item @qcode{"nofill"}
+## @item @qcode{"nofill"} (default)
 ## ILU factorization with no fill-in (ILU(0)).
 ##
 ## Additional supported options: @code{milu}.
@@ -52,7 +52,7 @@
 ##
 ## Additional supported options: @code{milu}, @code{droptol}.
 ##
-## @item @qcode{"ilutp"} (default)
+## @item @qcode{"ilutp"}
 ## ILU factorization with threshold and pivoting.
 ##
 ## Additional supported options: @code{milu}, @code{droptol}, @code{udiag},
@@ -137,7 +137,7 @@
 ## This shows that @var{A} has 7,840 nonzeros, the complete LU factorization
 ## has 126,478 nonzeros, and the incomplete LU factorization, with 0 level of
 ## fill-in, has 7,840 nonzeros, the same amount as @var{A}.  Taken from:
-## http://www.mathworks.com/help/matlab/ref/ilu.html
+## @url{http://www.mathworks.com/help/matlab/ref/ilu.html}
 ##
 ## @example
 ## @group
@@ -154,7 +154,8 @@
 ## @end example
 ##
 ## This example uses ILU as preconditioner for a random FEM-Matrix, which has a
-## large condition number.  Without @var{L} and @var{U} BICG would not converge.
+## large condition number.  Without @var{L} and @var{U} BICG would not
+## converge.
 ##
 ## @seealso{lu, ichol, bicg, gmres}
 ## @end deftypefn
@@ -166,11 +167,11 @@ function [L, U, P] = ilu (A, opts = struct ())
   endif
 
   if (! (issparse (A) && issquare (A)))
-    error ("ichol: A must be a sparse square matrix");
+    error ("ilu: A must be a sparse square matrix");
   endif
 
   if (! isstruct (opts))
-    error ("ichol: OPTS must be a structure.");
+    error ("ilu: OPTS must be a structure.");
   endif
 
   ## If A is empty then return empty L, U and P for Matlab compatibility
@@ -241,12 +242,12 @@ function [L, U, P] = ilu (A, opts = struct ())
           P = speye (length (A));
         endif
     case "ilutp"
-        if (nargout == 2)
-          [L, U]  = __ilutp__ (A, opts.droptol, opts.thresh,
-                                  opts.milu, opts.udiag);
-        elseif (nargout == 3)
-          [L, U, P]  = __ilutp__ (A, opts.droptol, opts.thresh,
-                                     opts.milu, opts.udiag);
+        if (nargout == 3)
+          [L, U, P] = __ilutp__ (A, opts.droptol, opts.thresh,
+                                    opts.milu, opts.udiag);
+        else
+          [L, U] = __ilutp__ (A, opts.droptol, opts.thresh,
+                                 opts.milu, opts.udiag);
         endif
   endswitch
 
@@ -261,9 +262,11 @@ endfunction
 %! n = 1600;
 %! dtol = 0.1;
 %! A = gallery ("neumann", n) + speye (n);
+
 %!test
 %! opts.type = "nofill";
 %! assert (nnz (ilu (A, opts)), 7840);
+
 ## This test has been verified in both Matlab and Octave.
 %!test
 %! opts.type = "crout";
@@ -312,7 +315,7 @@ endfunction
 
 ## Tests for real matrices of different sizes for ilu0, iluc and ilutp.
 ## The difference A - L*U should be not greater than eps because with droptol
-## equaling 0, the LU complete factorization is performed.
+## equal to 0, the LU complete factorization is performed.
 %!shared n_tiny, n_small, n_medium, n_large, A_tiny, A_small, A_medium, A_large
 %! n_tiny = 5;
 %! n_small = 40;
@@ -322,7 +325,7 @@ endfunction
 %! A_small = sprand (n_small, n_small, 1/n_small) + speye (n_small);
 %! A_medium = sprand (n_medium, n_medium, 1/n_medium) + speye (n_medium);
 %! A_large = sprand (n_large, n_large, 1/n_large/10) + speye (n_large);
-%!
+
 %!test
 %! opts.type = "nofill";
 %! [L, U] = ilu (A_tiny);
@@ -339,7 +342,7 @@ endfunction
 %! opts.type = "nofill";
 %! [L, U] = ilu (A_large);
 %! assert (norm (A_large - L*U, "fro") / norm (A_large, "fro"), 0, 1);
-%!
+
 %!test
 %! opts.type = "crout";
 %! [L, U] = ilu (A_tiny, opts);
@@ -356,7 +359,7 @@ endfunction
 %! opts.type = "crout";
 %! [L, U] = ilu (A_large, opts);
 %! assert (norm (A_large - L*U, "fro") / norm (A_large, "fro"), eps, eps);
-%!
+
 %!test
 %! opts.type = "ilutp";
 %! opts.droptol = 0;
@@ -396,7 +399,7 @@ endfunction
 %!   i * sprand (n_medium, n_medium, 1/n_medium) + speye (n_medium);
 %! A_large = sprand (n_large, n_large, 1/n_large/10) + ...
 %!   i * sprand (n_large, n_large, 1/n_large/10) + speye (n_large);
-%!
+
 %!test
 %! opts.type = "nofill";
 %! [L, U] = ilu (A_tiny);
@@ -413,7 +416,7 @@ endfunction
 %! opts.type = "nofill";
 %! [L, U] = ilu (A_large);
 %! assert (norm (A_large - L*U, "fro") / norm (A_large, "fro"), 0, 1);
-%!
+
 %!test
 %! opts.type = "crout";
 %! [L, U] = ilu (A_tiny, opts);
@@ -430,7 +433,7 @@ endfunction
 %! opts.type = "crout";
 %! [L, U] = ilu (A_large, opts);
 %! assert (norm (A_large - L*U, "fro") / norm (A_large, "fro"), eps, eps);
-%!
+
 %!test
 %! opts.type = "ilutp";
 %! opts.droptol = 0;
@@ -457,29 +460,40 @@ endfunction
 %! assert (norm (A_large - L*U, "fro") / norm (A_large, "fro"), eps, eps);
 
 ## Specific tests for ilutp
+%!shared A
+%! A = sparse ([0 0 4 3 1; 5 1 2.3 2 4.5; 0 0 0 2 1;0 0 8 0 2.2; 0 0 9 9 1 ]);
 
-%!shared a1, a2
-%! a1 = sparse ([0 0 4 3 1; 5 1 2.3 2 4.5; 0 0 0 2 1;0 0 8 0 2.2; 0 0 9 9 1 ]);
-%! a2 = sparse ([3 1 0 0 4; 3 1 0 0 -2;0 0 8 0 0; 0 4 0 4 -4.5; 0 -1 0 0 1]);
 %!test
 %! opts.udiag = 1;
 %! opts.type = "ilutp";
 %! opts.droptol = 0.2;
-%! [L, U, P] = ilu (a1, opts);
+%! [L, U, P] = ilu (A, opts);
 %! assert (norm (U, "fro"), 17.4577, 1e-4);
 %! assert (norm (L, "fro"), 2.4192, 1e-4);
+
+%!error <encountered a pivot equal to 0>
+%! opts.type = "ilutp";
 %! opts.udiag = 0;
-%! #fail ("ilu (a1, opts)");
-%!
-%!test
+%! opts.droptol = 0.2;
+%! ilu (A, opts);
+
+## Matlab R2017b doesn't error, but returns a singular L which isn't helpful.
+%!error <encountered a pivot equal to 0>
+%! A = sparse ([3 1 0 0 4; 3 1 0 0 -2;0 0 8 0 0; 0 4 0 4 -4.5; 0 -1 0 0 1]);
 %! opts.type = "ilutp";
 %! opts.droptol = 0;
 %! opts.thresh = 0;
 %! opts.milu = "row";
-%! #fail ("ilu (a2, opts)");
+%! [L, U, P] = ilu (A, opts);
+
+%!test <*53440>
+%! A = sparse (magic (4));
+%! opts.type = "ilutp";
+%! [L, U] = ilu (A, opts);
+%! assert (L * U, A, eps)
 
 ## Tests for input validation
-%!shared A_tiny
+%!shared A_tiny, opts
 %! A_tiny = spconvert ([1 4 2 3 3 4 2 5; 1 1 2 3 4 4 5 5; 1 2 3 4 5 6 7 8]');
 
 %!test
@@ -494,43 +508,61 @@ endfunction
 %! [L, U] = ilu (sparse ([]), opts);
 %! assert (isempty (L));
 %! assert (isempty (U));
+
 %!error <A must be a sparse square matrix> ilu (0)
 %!error <A must be a sparse square matrix> ilu ([])
 %!error <zero on the diagonal> ilu (sparse (0))
 
-%!test
+%!error <invalid TYPE specified>
 %! opts.type = "foo";
-%! fail ("ilu (A_tiny, opts)", "invalid TYPE specified");
+%! ilu (A_tiny, opts);
+%!error <invalid TYPE specified>
 %! opts.type = 1;
-%! fail ("ilu (A_tiny, opts)", "invalid TYPE specified");
+%! ilu (A_tiny, opts);
+%!error <invalid TYPE specified>
 %! opts.type = [];
-%! fail ("ilu (A_tiny, opts)", "invalid TYPE specified");
-%!test
-%! opts.droptol = -1;
-%! fail ("ilu (A_tiny, opts)", "DROPTOL must be a non-negative real scalar");
-%! opts.droptol = 0.5i;
-%! fail ("ilu (A_tiny, opts)", "DROPTOL must be a non-negative real scalar");
-%! opts.droptol = [];
-%! fail ("ilu (A_tiny, opts)", "DROPTOL must be a non-negative real scalar");
-%!test
-%! opts.milu = "foo";
-%! fail ("ilu (A_tiny, opts)", 'MILU must be one of "off"');
-%! opts.milu = 1;
-%! fail ("ilu (A_tiny, opts)", 'MILU must be one of "off"');
-%! opts.milu = [];
-%! fail ("ilu (A_tiny, opts)", 'MILU must be one of "off"');
-%!test
-%! opts.udiag = -1;
-%! fail ("ilu (A_tiny, opts)", "UDIAG must be 0 or 1");
-%! opts.udiag = 0.5i;
-%! fail ("ilu (A_tiny, opts)", "UDIAG must be 0 or 1");
-%! opts.udiag = [];
-%! fail ("ilu (A_tiny, opts)", "UDIAG must be 0 or 1");
-%!test
-%! opts.thresh = -1;
-%! fail ("ilu (A_tiny, opts)", "THRESH must be a scalar");
-%! opts.thresh = 0.5i;
-%! fail ("ilu (A_tiny, opts)", "THRESH must be a scalar");
-%! opts.thresh = [];
-%! fail ("ilu (A_tiny, opts)", "THRESH must be a scalar");
+%! ilu (A_tiny, opts);
 
+%!error <DROPTOL must be a non-negative real scalar>
+%! clear opts;
+%! opts.droptol = -1;
+%! ilu (A_tiny, opts);
+%!error <DROPTOL must be a non-negative real scalar>
+%! opts.droptol = 0.5i;
+%! ilu (A_tiny, opts);
+%!error <DROPTOL must be a non-negative real scalar>
+%! opts.droptol = [];
+%! ilu (A_tiny, opts);
+
+%!error <MILU must be one of "off", "col", or "row">
+%! clear opts;
+%! opts.milu = "foo";
+%! ilu (A_tiny, opts);
+%!error <MILU must be one of "off", "col", or "row">
+%! opts.milu = 1;
+%! ilu (A_tiny, opts);
+%!error <MILU must be one of "off", "col", or "row">
+%! opts.milu = [];
+%! ilu (A_tiny, opts);
+
+%!error <UDIAG must be 0 or 1>
+%! clear opts;
+%! opts.udiag = -1;
+%! ilu (A_tiny, opts);
+%!error <UDIAG must be 0 or 1>
+%! opts.udiag = 0.5i;
+%! ilu (A_tiny, opts);
+%!error <UDIAG must be 0 or 1>
+%! opts.udiag = [];
+%! ilu (A_tiny, opts);
+
+%!error <THRESH must be a scalar in the range \[0, 1\]>
+%! clear opts;
+%! opts.thresh = -1;
+%! ilu (A_tiny, opts);
+%!error <THRESH must be a scalar in the range \[0, 1\]>
+%! opts.thresh = 0.5i;
+%! ilu (A_tiny, opts);
+%!error <THRESH must be a scalar in the range \[0, 1\]>
+%! opts.thresh = [];
+%! ilu (A_tiny, opts);
